"""
This module implements meta-programming utilities.
They should be used only when absolutely necessary.
"""
import inspect

from pathlib import Path


def reference_caller_function(*, depth=1):
    """
    Returns a reference to the calling function, based on the specified depth.

    This function inspects the stack frames to retrieve the function that called it.
    The depth parameter specifies how many frames back to look:
        - a depth of 0 will return the function from which the call is made
        - a depth of 1 (the default) will return the immediate caller
        - a depth of 2 will return the caller's caller, and so on

    Parameters:
        depth (int, optional): The number of stack frames back to inspect to find the calling function. (default: 1)

    Returns:
        function: A reference to the function at the specified depth.

    Raises:
        TypeError: If depth is not an integer.
        ValueError: If depth is less than 0 or more than the number of available stack frames.

    Example:
        >>> def foo():
        ...     caller = reference_caller_function()
        ...     print(f'{caller.__name__}() called foo()')
        ...
        >>> def bar():
        ...     foo()
        ...
        >>> bar()
        bar() called foo()

    """
    # I have to execute the preconditions both here and in extract_call_stack_info() because
    # of the depth+1 operation. Otherwise, I might do `True + 1` which python evaluates to 2
    if not isinstance(depth, int):
        raise TypeError(f'depth must be an integer, not {type(depth).__name__}')
    elif isinstance(depth, bool):
        raise TypeError(f'depth must be an integer, not bool')
    elif depth < 0:
        raise ValueError(f'depth must be >= 0, not {depth}')

    # reference_caller_function() adds an extra frame to the stack so you should use depth=depth+1
    return extract_call_stack_info(depth=depth+1)[1]


def extract_call_stack_info(*, depth=0):
    """
    Returns the file path, caller function reference, and line number, based on the specified depth.

    This function inspects the stack frames to retrieve the function that called it.
    The depth parameter specifies how many frames back to look:
        - a depth of 0 will return the function from which extract_call_stack_info() was called
        - a depth of 1 (the default) will return the caller one level up
        - a depth of 2 will return the caller two levels up, and so on

    Parameters:
        depth (int, optional): The number of stack frames to go back. (default: 0)

    Returns:
        tuple: A tuple containing the file path, caller function reference, and line number.

    Raises:
        TypeError: If depth is not an integer.
        ValueError: If depth is less than 0 or more than the number of available stack frames.

    """
    if not isinstance(depth, int):
        raise TypeError(f'depth must be an integer, not {type(depth).__name__}')
    elif isinstance(depth, bool):
        raise TypeError(f'depth must be an integer, not bool')
    elif depth < 0:
        raise ValueError(f'depth must be >= 0, not {depth}')

    stack = inspect.stack()
    if depth >= len(stack):
        raise ValueError(f'You\'re getting ahead of yourself. There are only {len(stack)} frames in the stack, not {depth}.')

    # The stack is ordered from the most recent frame to the oldest frame
    # The most recent frame is this function, so we need to increase the depth by 1
    frame_info = stack[depth+1]
    caller = frame_info.frame.f_code.co_name
    func = _search(caller, frame_info.frame.f_globals, recursion_limit=1)

    return str(Path(frame_info.filename).resolve()), func, frame_info.lineno


def _search(key, data, recursion_limit):
    func = data.get(key)
    if func:
        return func

    next_level = []
    for k, obj in data.items():
        if k.startswith('__'):
            continue

        obj_2 = getattr(obj, '__dict__', {})
        if key in obj_2:
            return obj_2[key]

        if obj_2:
            next_level.append(obj_2)

    if recursion_limit == 0:
        return None

    for obj in next_level:
        func = _search(key, obj, recursion_limit - 1)
        if func:
            return func



"""
This module contains various utility functions for checking and comparing data.

- checking the type
- checking the structure
- checking if two collections share some values or properties

Examples:
    >>> contains_any_of([1, 2, 3], [1, True, None])
    True
    >>> is_sequence_of_dicts([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}])
    True
    >>> is_singleton(True)
    True
    >>> is_singleton(None)
    True
"""

import collections
import inspect
from decimal import Decimal
from fractions import Fraction


def contains_all_of(haystack, needles):
    """
    Check if all the needles are present in the haystack.

    Args:
        haystack (iterable):
            The iterable to search in.
        needles (iterable):
            The iterable containing elements to search for.

    Returns:
        bool:
            True if all elements in needles are found in the haystack, False otherwise.

    Raises:
        TypeError:
            If either haystack or needles is not an iterable.

    Examples:
        >>> contains_all_of([4, True, None, 1, 3], [1, True, 3])
        True
        >>> contains_all_of([1, 2, True], [1, 2, 3])
        False

    """
    if not is_iterable(haystack):
        raise TypeError(f'<haystack> expects an iterable. Got {type(haystack).__name__}: {haystack!r}')
    elif not is_iterable(needles):
        raise TypeError(f'<needles> expects an iterable. Got {type(needles).__name__}: {needles!r}')

    for needle in needles:
        if not is_any_of(needle, haystack):
            return False

    return True


def contains_any_of(haystack, needles):
    """
    Check if any needle is present in the haystack.

    Args:
        haystack (iterable):
            The iterable to search in.
        needles (iterable):
            The iterable containing elements to search for.

    Returns:
        bool:
            True if at least one element in needles is found in the haystack, False otherwise.

    Raises:
        TypeError:
            If either haystack or needles is not an iterable.

    Examples:
        >>> contains_any_of([1, 2, 3], [1, True, None])
        True
        >>> contains_any_of([1, True, 3], [4, True, None])
        True
        >>> contains_any_of([1, 2, 3], [4, True, None])
        False

    """
    if not is_iterable(haystack):
        raise TypeError(f'<haystack> expects an iterable. Got {type(haystack).__name__}: {haystack!r}')
    elif not is_iterable(needles):
        raise TypeError(f'<needles> expects an iterable. Got {type(needles).__name__}: {needles!r}')

    for needle in needles:
        if is_any_of(needle, haystack):
            return True

    return False


def contains_none_of(haystack, needles):
    """
    Check if none of the needles are present in the haystack.

    Args:
        haystack (iterable):
            The iterable to search in.
        needles (iterable):
            The iterable containing elements to search for.

    Returns:
        bool:
            True if none of the elements in needles are found in the haystack, False otherwise.

    Examples:
        >>> contains_none_of([1, 2, 3], [4, True, None])
        True
        >>> contains_none_of([1, 2, 3], [1, True, None])
        False
        >>> contains_none_of([1, True, 3], [4, True, None])
        False

    """
    return not contains_any_of(haystack, needles)


def dicts_share_key_order(dict_1, dict_2, recursive=False):
    """
    Check if two dictionaries share the same key order.

    Args:
        dict_1 (dict):
            The first dictionary.
        dict_2 (dict):
            The second dictionary.
        recursive (bool, optional):
            Specifies whether to recursively compare nested dictionaries. Defaults to False.

    Returns:
        bool:
            True if the dictionaries share the same key order, False otherwise.

    Raises:
        TypeError:
            If recursive is not a boolean value or if the types of dict_1 and dict_2 are not compatible.

    Examples:
        >>> dicts_share_key_order({'a': 1, 'b': 2}, {'a': 3, 'b': 4})
        True
        >>> dicts_share_key_order({'a': 1, 'b': 2}, {'c': 1, 'd': 2})
        False
        >>> dicts_share_key_order({'a': 1, 'b': 2}, {'c': 1})
        False

    """
    if not isinstance(recursive, bool):
        raise TypeError(f'<dict_2> must be True or False. Got {type(recursive)}')

    if isinstance(dict_1, dict):
        if not isinstance(dict_2, dict):
            raise TypeError(f'<dict_2> must be a dict. Got {type(dict_2)}')

        if list(dict_1.keys()) != list(dict_2.keys()):
            return False

        if recursive:
            return all(
                dicts_share_key_order(v_1, v_2, recursive=recursive)
                for v_1, v_2 in zip(dict_1.values(), dict_2.values()))
        else:
            return True
    elif recursive and is_sequence(dict_1):
        if not is_sequence(dict_2):
            raise TypeError(f'<dict_2> must be a sequence. Got {type(dict_2)}')

        return all(
            dicts_share_key_order(d_1, d_2, recursive=recursive)
            for d_1, d_2 in zip(dict_1, dict_2))
    elif not recursive:
        raise TypeError(f'<dict_1> and <dict_2> must be dicts. Got {type(dict_1)} and {type(dict_2)}')
    else:
        return True


def dicts_share_value_order(dict_1, dict_2, recursive=False):
    """
    Check if two dictionaries share the same value order.

    Args:
        dict_1 (dict):
            The first dictionary.
        dict_2 (dict):
            The second dictionary.
        recursive (bool, optional):
            Specifies whether to recursively compare nested dictionaries. Defaults to False.

    Returns:
        bool:
            True if the dictionaries share the same value order, False otherwise.

    Raises:
        TypeError:
            If recursive is not a boolean value or if the types of dict_1 and dict_2 are not compatible.

    Examples:
        >>> dicts_share_value_order({'a': 1, 'b': 2}, {'c': 1, 'd': 2})
        True
        >>> dicts_share_value_order({'a': 1, 'b': 2}, {'c': 1, 'd': 3})
        False
        >>> dicts_share_value_order({'a': 1, 'b': 2}, {'c': 1})
        False

    """
    if not isinstance(recursive, bool):
        raise TypeError(f'<dict_2> must be True or False. Got {type(recursive)}')

    if isinstance(dict_1, dict):
        if not isinstance(dict_2, dict):
            raise TypeError(f'<dict_2> must be a dict. Got {type(dict_2)}')

        if list(dict_1.values()) != list(dict_2.values()):
            return False

        if recursive:
            return all(
                dicts_share_value_order(v_1, v_2, recursive=recursive)
                for v_1, v_2 in zip(dict_1.values(), dict_2.values()))
        else:
            return True
    elif recursive and is_sequence(dict_1):
        return all(
            dicts_share_value_order(d_1, d_2, recursive=recursive)
            for d_1, d_2 in zip(dict_1, dict_2))
    elif not recursive:
        raise TypeError(f'<dict_1> and <dict_2> must be dicts. Got {type(dict_1)} and {type(dict_2)}')
    else:
        return True


def is_any_of(needle, haystack):
    """
    Check if the needle is in the haystack. This correctly compares singletons (None, True, False)
    using `is` and other values using `==`.

    Args:
        needle (Any):
            The value to compare against the elements in the haystack.
        haystack (Iterable):
            The iterable containing the elements to compare with the needle.

    Returns:
        bool:
            True if the needle is equal to any element in the haystack, False otherwise.

    Raises:
        TypeError:
            If the haystack is not an iterable.

    Examples:
        >>> is_any_of(1, [1, 2, 3])
        True
        >>> is_any_of(0, [1, 2, 3])
        False
        >>> is_any_of(True, [1, 2, 3])
        False

    """
    if not is_iterable(haystack):
        raise TypeError(f'<haystack> expects an iterable. Got {type(haystack).__name__}: {haystack!r}')

    return any(is_equal(needle, x) for x in haystack)


def is_dict_of_sequences(x, *, empty_ok=True):
    """
    Check if the input is a dictionary where all values are sequences.

    Args:
        x (Any): The input to check.
        empty_ok (bool, optional): Specifies whether an empty dictionary is allowed. Defaults to True.

    Returns:
        bool: True if the input is a dictionary and all values are sequences,
            False otherwise.

    Examples:
        >>> is_dict_of_sequences({'a': [1, 2], 'b': [3, 4]})
        True
        >>> is_dict_of_sequences({'a': [1, 2], 'b': [3, 4, 5]})
        True
        >>> is_dict_of_sequences({'a': [1, 2], 'b': 100})
        False
        >>> is_dict_of_sequences([{'a': 1}, {'a': 2}])
        False
        >>> is_dict_of_sequences({}, empty_ok=False)
        False
        >>> is_dict_of_sequences({}, empty_ok=True)
        True

    """
    if not isinstance(empty_ok, bool):
        raise TypeError(f'<empty_ok> expects False or True. Got {type(empty_ok)}')

    if isinstance(x, dict):
        if len(x) >= 1 and all(is_sequence(v) for v in x.values()):
            return True
        elif len(x) == 0 and empty_ok:
            return True

    return False


def is_dict_of_sequences_uniform(x, *, empty_ok=True):
    """
    Check if the input is a dictionary where all values are sequences of the same length.

    Args:
        x (Any): The input to check.
        empty_ok (bool, optional): Specifies whether an empty dictionary is allowed. Defaults to True.

    Returns:
        bool: True if the input is a dictionary and all values are sequences of the same length,
            False otherwise.

    Examples:
        >>> is_dict_of_sequences_uniform({'a': [1, 2], 'b': [3, 4]})
        True
        >>> is_dict_of_sequences_uniform({'a': [], 'b': []})
        True
        >>> is_dict_of_sequences_uniform({'a': [1, 2], 'b': [3]})
        False
        >>> is_dict_of_sequences_uniform({'a': [1, 2], 'b': [3, 4, 5]})
        False
        >>> is_dict_of_sequences_uniform({}, empty_ok=False)
        False
        >>> is_dict_of_sequences_uniform({}, empty_ok=True)
        True

    """
    if not is_dict_of_sequences(x, empty_ok=empty_ok):
        return False
    elif len(x) == 0 and empty_ok:
        return True

    n = len(list(x.values())[0])
    return all(len(v) == n for v in x.values())


def is_equal(a, b):
    """
    Check if two objects are equal. This correctly compares singletons
    (None, True, False) using `is` and other values using `==`.

    Args:
        a (Any): The first object.
        b (Any): The second object.

    Returns:
        bool: True if the objects are equal, False otherwise.

    Examples:
        >>> is_equal(1, 1)
        True
        >>> is_equal(False, False)
        True
        >>> is_equal(1, True)
        False
        >>> is_equal(1, 2)
        False

    """
    if is_singleton(a) or is_singleton(b):
        return a is b
    else:
        return a == b


def is_generator(x):
    # noinspection PyUnresolvedReferences
    """
    Check if an object is a generator expression or generator function.

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is a generator, False otherwise.

    Examples:
        >>> def _gen(): yield 1
        >>> is_generator(_gen)
        True
        >>> is_generator(i for i in range(5))
        True
        >>> is_generator(None)
        False
        >>> is_generator(1)
        False

    """
    return is_generator_expression(x) or is_generator_function(x)


def is_generator_expression(x):
    # noinspection PyUnresolvedReferences
    """
    Check if an object is a generator expression.

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is a generator expression, False otherwise.

    Examples:
        >>> is_generator_expression(i for i in range(5))
        True
        >>> def _gen(): yield 1
        >>> is_generator_expression(_gen)
        False
        >>> is_generator_expression(None)
        False
        >>> is_generator_expression(1)
        False

    """
    return inspect.isgenerator(x)


def is_generator_function(x):
    # noinspection PyUnresolvedReferences
    """
    Check if an object is a generator function.

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is a generator function, False otherwise.

    Examples:
        >>> def _gen(): yield 1
        >>> is_generator_function(_gen)
        True
        >>> is_generator_function(i for i in range(5))
        False
        >>> is_generator_function(None)
        False
        >>> is_generator_function(1)
        False

    """
    return inspect.isgeneratorfunction(x)


def is_iterable(x):
    """
    Check if an object is iterable.

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is iterable, False otherwise.

    Examples:
        >>> is_iterable([1,2,3])
        True
        >>> is_iterable(map(float, range(5)))
        True
        >>> is_iterable(None)
        False
        >>> is_iterable(1)
        False
        >>> is_iterable('hello')
        False

    """
    if is_generator(x):
        return True

    try:
        # The only reliable way to determine whether an object is iterable is to call iter(obj).
        iter(x)
        # str, bytes, bytearray are theoretically a iterables,
        # but in practice we use them as primitives
        return not isinstance(x, (str, bytes, bytearray))
    except TypeError:
        return False


def is_iterable_or_str(x):
    """
    Check if an object is iterable or a string.

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is iterable or a string, False otherwise.

    Examples:
        >>> is_iterable_or_str([1,2,3])
        True
        >>> is_iterable_or_str('hello')
        True
        >>> is_iterable_or_str(map(float, range(5)))
        True
        >>> is_iterable_or_str(None)
        False
        >>> is_iterable_or_str(1)
        False

    """
    return is_generator(x) or isinstance(x, collections.abc.Iterable)


def is_iterator(x):
    """
    Check if an object is an iterator.

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is an iterator, False otherwise.

    Examples:
        >>> is_iterator(iter([]))
        True
        >>> is_iterator(iter({1,2,3}))
        True
        >>> is_iterator(map(float, range(5)))
        True
        >>> is_iterator(None)
        False
        >>> is_iterator([])
        False

    """
    return is_generator(x) or isinstance(x, collections.abc.Iterator)


def is_none_of(needle, haystack):
    """
    Check if the needle is not in the haystack. This correctly compares
    singletons (None, True, False) using `is` and other values using `==`.

    Args:
        needle (Any): The value to compare against the elements in the haystack.
        haystack (Iterable): The iterable containing the elements to compare with the needle.

    Returns:
        bool: True if the needle is not equal to any element in the haystack, False otherwise.

    Examples:
        >>> is_none_of(1, [])
        True
        >>> is_none_of(2, [1, 3])
        True
        >>> is_none_of(1, [1])
        False
        >>> is_none_of([1, 2], [[1, 2], 3, 4])
        False

    """
    return not is_any_of(needle, haystack)


def is_number(x):
    """
    Check if an object is a number (int, float, complex, Decimal, Fraction).

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is a number, False otherwise.

    Examples:
        >>> is_number(1)
        True
        >>> is_number(Decimal('3.14'))
        True
        >>> is_number(-0.1)
        True
        >>> is_number(None)
        False
        >>> is_number([])
        False

    """
    if isinstance(x, bool):
        return None

    return isinstance(x, (int, float, complex, Decimal, Fraction))


def is_primitive(x):
    """
    Check if an object is a primitive.
    Return True for str because in practice we use it as a primitive,
    even though it is theoretically a collection

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is a primitive, False otherwise.

    Examples:
        >>> is_primitive(1)
        True
        >>> is_primitive('hello')
        True
        >>> is_primitive(False)
        True
        >>> is_primitive([])
        False

    """
    return isinstance(x, (str, int, float, bool, type(None), bytes, bytearray))


def is_sequence(x):
    """
    Check if an object is a sequence. This function excludes str, bytes,
    and bytearray from being considered sequences, although because
    in practice they are used as primitives

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is a sequence, False otherwise.

    Examples:
        >>> is_sequence([1, 2, 3])
        True
        >>> is_sequence('hello')
        False
        >>> is_sequence(0)
        False
        >>> is_sequence({1, 2, 3})
        False
        >>> is_sequence({'a': 1})
        False

    """
    return isinstance(x, collections.abc.Sequence) and not isinstance(x, (str, bytes, bytearray))


def is_sequence_of_dicts(x, *, empty_ok=True):
    """
    Check if an object is a sequence of dictionaries.

    Args:
        x (Any): The object to check.
        empty_ok (bool, optional): Specifies whether an empty sequence is allowed. Defaults to True.

    Returns:
        bool: True if the object is a sequence of dictionaries, False otherwise.

    Examples:
        >>> is_sequence_of_dicts([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}])
        True
        >>> is_sequence_of_dicts([{}, {}])
        True
        >>> is_sequence_of_dicts([{'a': 1, 'b': 2}, {'a': 3, 'b': 4, 'c': 5}])
        True
        >>> is_sequence_of_dicts([], empty_ok=False)
        False
        >>> is_sequence_of_dicts([], empty_ok=True)
        True

    """
    if not isinstance(empty_ok, bool):
        raise TypeError(f'<empty_ok> expects False or True. Got {type(empty_ok)}')

    if is_sequence(x):
        if len(x) >= 1 and all(isinstance(xi, dict) for xi in x):
            return True
        elif len(x) == 0 and empty_ok:
            return True

    return False


def is_sequence_of_dicts_uniform(x, *, empty_ok=True):
    """
    Check if x is a sequence of dictionaries and all dicts have the same keys.

    Args:
        x (Any): The sequence of dictionaries to check.
        empty_ok (bool, optional): Specifies whether an empty sequence is allowed. Defaults to True.

    Returns:
        bool: True if the object is a sequence of dictionaries and they have the same keys, False otherwise.

    Examples:
        >>> is_sequence_of_dicts_uniform([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}])
        True
        >>> is_sequence_of_dicts_uniform([{}, {}])
        True
        >>> is_sequence_of_dicts_uniform([{'a': 1, 'b': 2}, {'a': 3, 'b': 4, 'c': 5}])
        False
        >>> is_sequence_of_dicts_uniform([], empty_ok=False)
        False
        >>> is_sequence_of_dicts_uniform([], empty_ok=True)
        True

    """
    if not is_sequence_of_dicts(x, empty_ok=empty_ok):
        return False
    elif len(x) == 0 and empty_ok:
        return True

    keys = set(x[0].keys())
    return all(set(xi.keys()) == keys for xi in x)


def is_sequence_of_sequences(x, *, empty_ok=True):
    """
    Check if an object is a sequence of sequences.

    Args:
        x (Any): The object to check.
        empty_ok (bool, optional): Specifies whether an empty sequence is allowed. Defaults to True.

    Returns:
        bool: True if the object is a sequence of sequences, False otherwise.

    Examples:
        >>> is_sequence_of_sequences([[1, 2], [3, 4]])
        True
        >>> is_sequence_of_sequences([[], []])
        True
        >>> is_sequence_of_sequences([[1, 2], [3]])
        True
        >>> is_sequence_of_sequences([], empty_ok=False)
        False
        >>> is_sequence_of_sequences([], empty_ok=True)
        True

    """
    if not isinstance(empty_ok, bool):
        raise TypeError(f'<empty_ok> expects False or True. Got {type(empty_ok)}')

    if is_sequence(x):
        if len(x) >= 1 and all(is_sequence(xi) for xi in x):
            return True
        elif len(x) == 0 and empty_ok:
            return True

    return False


def is_sequence_of_sequences_uniform(x, *, empty_ok=True):
    """
    Check if x is a sequence of sequences and all sequences have the same length.

    Args:
        x (Any): The sequence of sequences to check.
        empty_ok (bool, optional): Specifies whether an empty sequence is allowed. Defaults to True.

    Returns:
        bool: True if the object is a sequence of sequences and they have the same length, False otherwise.

    Examples:
        >>> is_sequence_of_sequences_uniform([[1, 2], [3, 4]])
        True
        >>> is_sequence_of_sequences_uniform([[], []])
        True
        >>> is_sequence_of_sequences_uniform([[1, 2], [3]])
        False
        >>> is_sequence_of_sequences_uniform([], empty_ok=False)
        False
        >>> is_sequence_of_sequences_uniform([], empty_ok=True)
        True

    """
    if not is_sequence_of_sequences(x, empty_ok=empty_ok):
        return False
    elif len(x) == 0 and empty_ok:
        return True

    n = len(x[0])
    return all(len(xi) == n for xi in x)


def is_sequence_or_str(x):
    """
    Check if an object is a sequence or a string.

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is a sequence or a string, False otherwise.

    Examples:
        >>> is_sequence_or_str([1, 2, 3])
        True
        >>> is_sequence_or_str('hello')
        True
        >>> is_sequence_or_str(0)
        False
        >>> is_sequence_or_str({1, 2, 3})
        False
        >>> is_sequence_or_str({'a': 1})
        False

    """
    return isinstance(x, collections.abc.Sequence)


def is_singleton(x):
    """
    Check if an object is a singleton (True, False, or None).

    Args:
        x (Any): The object to check.

    Returns:
        bool: True if the object is a singleton, False otherwise.

    Examples:
        >>> is_singleton(True)
        True
        >>> is_singleton(None)
        True
        >>> is_singleton([None])
        False
        >>> is_singleton({'a': 1})
        False

    """
    return isinstance(x, (bool, type(None)))


if __name__ == '__main__':  # pragma: no cover
    import doctest
    doctest.testmod()



import pytest

# noinspection PyProtectedMember
from katalytic._pkg import _check, find_functions_marked_with, get_modules, get_version, mark
from katalytic.katalytic import __version__


class Test_check:
    def test_all(self):
        assert _check(None, '') is False
        assert _check('load', 'save') is False
        assert _check('load', 'save::*') is False
        assert _check('save::txt', 'load::*') is False
        assert _check('load::txt', 'load::') is False

        assert _check('load::txt', 'load::txt') is True
        assert _check('load::txt', 'load::*') is True
        assert _check('load::image::png', 'load::*') is True


class Test_mark:
    def test_internals(self):
        def f1(): pass

        @mark('_test_group-1')
        def f2(): pass

        @mark('_test_group-2')
        @mark('_test_group-1')
        def f3(): pass

        assert not hasattr(f1, '__katalytic_marks__')
        assert f2.__katalytic_marks__ == ('_test_group-1', )
        assert f3.__katalytic_marks__ == ('_test_group-2', '_test_group-1')

    @pytest.mark.parametrize('group', [0, None, {}, True])
    def test_TypeError(self, group):
        with pytest.raises(TypeError):
            # noinspection PyTypeChecker
            @mark(group)
            def f1(): pass

    @pytest.mark.parametrize('group', ['', ' ', ' \n ', '\t \t', 'no \t tabs', 'no \n newlines'])
    def test_ValueError(self, group):
        with pytest.raises(ValueError):
            @mark(group)
            def f1(): pass


class Test_get_functions_in_group:
    def test_empty(self):
        assert find_functions_marked_with('__not-used') == []
        assert find_functions_marked_with('__test_3') == []

    def test_pattern(self):
        found = find_functions_marked_with('__test_3*')
        found = [(name, groups) for name, _, groups in found]
        assert found == [
            ('__test', ['__test_300']),
            ('__test_2', ['__test_3::a', '__test_3::b'])
        ]

    def test_exact(self):
        found = find_functions_marked_with('__test_1')
        found = [(name, groups) for name, _, groups in found]
        assert found == [('__test', ['__test_1'])]

        found = find_functions_marked_with('__test_2')
        found = [(name, groups) for name, _, groups in found]
        assert found == [
            ('__test', ['__test_2']),
            ('__test_2', ['__test_2'])
        ]


class Test_get_modules:
    def test_all(self):
        modules = get_modules()
        assert 'katalytic._pkg' in [m.__name__ for m in modules]
        assert all(m.__name__.startswith('katalytic') for m in modules)


class Test_version:
    def test_is_ok(self):
        v, v_info = get_version('katalytic')

        assert __version__ is not None
        assert v is not None
        assert v_info is not None
        assert v == '.'.join(str(i) for i in v_info)
        assert v == __version__



from statistics import mean

from katalytic.data import pop_max
from katalytic.data.checks import is_sequence


def calc_bbox_area(bbox, fmt):
    """
    Calculate the area of a bounding box.

    Args:
        bbox (sequence): The bounding box whose area is to be calculated. It should be in the format specified by 'fmt'.
        fmt (str): The format of the bounding box. It could be one of the formats defined in `_FORMATS`.

    Returns:
        float: The area of the bounding box.

    """
    bbox = convert_bbox(bbox, fmt, 'xy_wh') if fmt != 'xy_wh' else bbox
    (_, _), (w, h) = bbox
    return w * h


def calc_bbox_center(bbox, fmt, *, as_int=False):
    """
    Calculate the center of a bounding box.

    Args:
        bbox (sequence): The bounding box whose center is to be calculated. It should be in the format specified by 'fmt'.
        fmt (str): The format of the bounding box. It could be one of the formats defined in `_FORMATS`.
        as_int (bool, optional): If True, the center coordinates are returned as integers. Defaults to False.

    Returns:
        tuple:
            The center of the bounding box as a tuple of two coordinates (x, y).
            If 'as_int' is True, the coordinates are integers, otherwise they are floats.
    """
    (x, y), (X, Y) = convert_bbox(bbox, fmt, 'xy_XY')
    cx, cy = (mean([x, X]), mean([y, Y]))

    if as_int:
        return (int(cx), int(cy))
    else:
        return (cx, cy)


def calc_IoB(bbox_1, fmt_1, bbox_2, fmt_2):
    """
    Calculate the Intersection over Bounding-box (IoB) between two bounding boxes, which is defined as the area of the intersection of `bbox_1` and `bbox_2` divided by the area of `bbox_1`.

    Args:
        bbox_1 (sequence): The first bounding box, defined in the format specified by `fmt_1`.
        fmt_1 (str): The format of the first bounding box. This defines how the coordinates in `bbox_1` are interpreted.
        bbox_2 (sequence): The second bounding box, defined in the format specified by `fmt_2`.
        fmt_2 (str): The format of the second bounding box. This defines how the coordinates in `bbox_2` are interpreted.

    Returns:
        float: The IoB between `bbox_1` and `bbox_2` If they don't intersect, returns 0.
    """
    intersection = intersect_bboxes(bbox_1, fmt_1, bbox_2, fmt_2)
    if intersection is None:
        return 0

    inter_area = calc_bbox_area(intersection, fmt_1)
    return inter_area / calc_bbox_area(bbox_1, fmt_1)


def calc_IoB_max(bbox_1, fmt_1, bbox_2, fmt_2):
    """
    Return the largest of (Intersection over bbox_1) and (Intersection over bbox_2).

    Args:
        bbox_1 (sequence): The first bounding box, defined in the format specified by `fmt_1`.
        fmt_1 (str): The format of the first bounding box. This defines how the coordinates in `bbox_1` are interpreted.
        bbox_2 (sequence): The second bounding box, defined in the format specified by `fmt_2`.
        fmt_2 (str): The format of the second bounding box. This defines how the coordinates in `bbox_2` are interpreted.

    Returns:
        float: The largest IoB value between `bbox_1` and `bbox_2`.
    """
    return max(
        calc_IoB(bbox_1, fmt_1, bbox_2, fmt_2),
        calc_IoB(bbox_2, fmt_2, bbox_1, fmt_1)
    )


def calc_IoB_min(bbox_1, fmt_1, bbox_2, fmt_2):
    """
    Return the smallest of (Intersection over bbox_1) and (Intersection over bbox_2).

    Args:
        bbox_1 (sequence): The first bounding box, defined in the format specified by `fmt_1`.
        fmt_1 (str): The format of the first bounding box. This defines how the coordinates in `bbox_1` are interpreted.
        bbox_2 (sequence): The second bounding box, defined in the format specified by `fmt_2`.
        fmt_2 (str): The format of the second bounding box. This defines how the coordinates in `bbox_2` are interpreted.

    Returns:
        float: The smallest IoB (Intersection over Bounding-box) value between `bbox_1` and `bbox_2`.
    """
    return min(
        calc_IoB(bbox_1, fmt_1, bbox_2, fmt_2),
        calc_IoB(bbox_2, fmt_2, bbox_1, fmt_1)
    )


def calc_IoU(bbox_1, fmt_1, bbox_2, fmt_2):
    """
    Calculate the Intersection over Union (IoU) between two bounding boxes, defined as the area of their intersection divided by the area of their union

    Args:
        bbox_1 (sequence): The first bounding box, defined in the format specified by `fmt_1`.
        fmt_1 (str): The format of the first bounding box. This defines how the coordinates in `bbox_1` are interpreted.
        bbox_2 (sequence): The second bounding box, defined in the format specified by `fmt_2`.
        fmt_2 (str): The format of the second bounding box. This defines how the coordinates in `bbox_2` are interpreted.

    Returns:
        float: The IoU of `bbox_1` and `bbox_2`. Returns 0 if `bbox_1` and `bbox_2` do not intersect.
    """
    intersection = intersect_bboxes(bbox_1, fmt_1, bbox_2, fmt_2)
    if intersection is None:
        return 0

    a1 = calc_bbox_area(bbox_1, fmt_1)
    a2 = calc_bbox_area(bbox_2, fmt_2)
    inter_area = calc_bbox_area(intersection, fmt_1)

    # the overlapping area is counted twice. Subtract it once
    union_area = a1 + a2 - inter_area
    return inter_area / union_area


def non_max_suppression(bboxes, format, *, max_IoU=0.5, scores=None):
    """
    Apply non-maximum suppression on a sequence of bounding boxes.

    Args:
        bboxes (list): A list of bounding boxes to perform non-maximum suppression on.
        format (str): The format in which the bounding boxes are specified.
        max_IoU (float, optional): The maximum IoU (Intersection over Union) value to consider a bounding box as a duplicate. Defaults to 0.5.
        scores (list, optional): A list of scores associated with each bounding box. If provided, must be the same length as `bboxes`.

    Raises:
        ValueError: If the format doesn't contain scores, and no `scores` argument is provided.

    Returns:
        list: The list of remaining bounding boxes after non-maximum suppression has been applied.
    """
    if not bboxes:
        return bboxes
    elif not format.endswith('s'):
        if scores:
            bboxes = set_bbox_scores(bboxes, scores, format, 'xyXYs')
        else:
            raise ValueError(
                'If the format doesn\'t contain scores, either change the call '
                'to non_max_suppression(..., scores=...) or use set_bbox_scores().')

    if format.endswith('s') and format != 'xyXYs':
        # standardise the format to simplify the implementation
        bboxes = [convert_bbox(bbox, format, 'xyXYs') for bbox in bboxes]

    kept, bboxes = pop_max(bboxes, condition=lambda bbox: bbox[-1])

    # remove bboxes with IoU > max_IoU than the kept bbox
    bboxes = [b for b in bboxes if calc_IoU(b, 'xyXYs', kept, 'xyXYs') < max_IoU]

    # call recursively on the remaining bboxes
    kept = [kept, *non_max_suppression(bboxes, 'xyXYs', max_IoU=max_IoU)]

    if format != 'xyXYs':
        # convert to the original format
        kept = [convert_bbox(bbox, 'xyXYs', format) for bbox in kept]

    return kept


def set_bbox_scores(bboxes, scores, before, after):
    """
    Insert or change the existing scores of a list of bounding boxes.

    Args:
        bboxes (list): A list of bounding boxes to set scores for.
        scores (list): A list of scores to set for each bounding box. Must have the same length as `bboxes`.
        before (str): The format in which the bounding boxes are currently specified.
        after (str): The desired format for the bounding boxes after the scores have been set. This format must end with 's'.

    Raises:
        ValueError: If `bboxes` and `scores` don't have the same length, if the `after` format doesn't end with 's', or if any score is not within the range [0, 1].

    Returns:
        list: The list of bounding boxes with scores set, in the specified `after` format.
    """
    if len(bboxes) != len(scores):
        raise ValueError('bboxes and scores must have the same length')
    elif not after.endswith('s'):
        raise ValueError('after format must be one of xyXYs, xyXY_s, xy_XY_s, xywhs, xywh_s, xy_wh_s')
    elif not all(0 <= s <= 1 for s in scores):
        raise ValueError('scores must be in the range [0, 1]')

    if before != 'xyXY':
        bboxes = [convert_bbox(bbox, before, 'xyXY') for bbox in bboxes]

    bboxes = [(*bbox, s) for bbox, s in zip(bboxes, scores)]
    if after == 'xyXYs':
        return bboxes
    else:
        return [convert_bbox(bbox, 'xyXYs', after) for bbox in bboxes]


def convert_bbox(bbox, before, after):
    """
    Convert a bounding box from one format to another.

    Args:
        bbox (sequence): The bounding box to convert. This should be in the format specified by `before`.
        before (str): The current format of the bounding box. This must be one of the recognized formats.
        after (str): The desired format for the bounding box. This must be one of the recognized formats.

    Raises:
        ValueError: If `before` or `after` are not recognized formats, if `bbox` is not in the `before` format, if `before` and `after` are the same, or if attempting to convert to a format that includes scores from a format that does not.

    Returns:
        sequence: The bounding box, converted to the `after` format.
    """
    if before not in _FORMATS:
        raise ValueError(f'Unknown before format {before!r}')
    elif after not in _FORMATS:
        raise ValueError(f'Unknown after format {after!r}')
    elif not is_bbox(bbox, before):
        raise ValueError(f'{bbox!r} is not in the {before!r} format')
    elif before == after:
        # I could just return the bbox, but I prefer to fail early
        # so the dev knows there might be a bug in his code
        # It makes the easy thing harder and the hard thing (debugging) easy
        raise ValueError(f'before and after formats are the same: {before!r}')
    elif after.endswith('s') and not before.endswith('s'):
        raise ValueError(f'Cannot convert from {before} to {after}')

    x = y = X = Y = w = h = s = None
    if before == 'xywh':      x,y,w,h = bbox
    elif before == 'xywhs':   x,y,w,h,s = bbox
    elif before == 'xywh_s':  (x,y,w,h), s = bbox
    elif before == 'xy_wh':   (x,y), (w,h) = bbox
    elif before == 'xy_wh_s': (x,y), (w,h), s = bbox
    elif before == 'xyXY':    x,y,X,Y = bbox
    elif before == 'xyXYs':   x,y,X,Y,s = bbox
    elif before == 'xyXY_s':  (x,y,X,Y), s = bbox
    elif before == 'xy_XY':   (x,y), (X,Y) = bbox
    elif before == 'xy_XY_s': (x,y), (X,Y), s = bbox

    if X is None:
        X = x + w
        Y = y + h
    elif w is None:
        w = X - x
        h = Y - y

    if after == 'xywh':      return x,y,w,h
    elif after == 'xywhs':   return x,y,w,h,s
    elif after == 'xywh_s':  return (x,y,w,h), s
    elif after == 'xy_wh':   return (x,y), (w,h)
    elif after == 'xy_wh_s': return (x,y), (w,h), s
    elif after == 'xyXY':    return x,y,X,Y
    elif after == 'xyXYs':   return x,y,X,Y,s
    elif after == 'xyXY_s':  return (x,y,X,Y), s
    elif after == 'xy_XY':   return (x,y), (X,Y)
    elif after == 'xy_XY_s': return (x,y), (X,Y), s


def intersect_bboxes(bbox_1, fmt_1, bbox_2, fmt_2):
    """
    Calculate the intersection of two bounding boxes.

    Args:
        bbox_1 (sequence): The first bounding box.
        fmt_1 (str): The format of the first bounding box.
        bbox_2 (sequence): The second bounding box.
        fmt_2 (str): The format of the second bounding box.

    Returns:
        list or None: The intersection of the two bounding boxes in the format of `fmt_1`, or `None` if the bounding boxes do not intersect.
    """
    if fmt_1.endswith('s'):
        s = bbox_1[-1]

    bbox_1 = convert_bbox(bbox_1, fmt_1, 'xy_XY') if fmt_1 != 'xy_XY' else bbox_1
    bbox_2 = convert_bbox(bbox_2, fmt_2, 'xy_XY') if fmt_2 != 'xy_XY' else bbox_2

    (x1, y1), (X1, Y1) = bbox_1
    (x2, y2), (X2, Y2) = bbox_2

    x3, y3 = max(x1, x2), max(y1, y2)
    X3, Y3 = min(X1, X2), min(Y1, Y2)

    if (0 <= x3 < X3) and (0 <= y3 < Y3):
        bbox_3 = [(x3, y3), (X3, Y3)]
        if fmt_1.endswith('s'):
            # noinspection PyUnboundLocalVariable
            bbox_3 = convert_bbox([*bbox_3, s], 'xy_XY_s', fmt_1)
        elif fmt_1 != 'xy_XY':
            bbox_3 = convert_bbox(bbox_3, 'xy_XY', fmt_1)

        return type(bbox_1)(bbox_3)
    else:
        return None


def is_bbox(bbox, fmt):
    """
    Check if the provided bounding box is of the specified format.

    Args:
        bbox (sequence): The bounding box to be checked.
        fmt (str): The format to check against.

    Returns:
        bool: `True` if the bounding box is of the specified format, `False` otherwise.

    Raises:
        ValueError: If the specified format is not recognized.
    """
    if fmt not in _FORMATS:
        available = ", ".join(_FORMATS)
        raise ValueError(f'Unknown format {fmt!r}. Available formats: {available}')

    return _FORMATS[fmt](bbox)


def _is_xy_wh(bbox):
    """WARNING:
    Returns True if the bbox has the xy_XY format
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y), (w, h) = bbox
        return (x >= 0) and (y >= 0) and (w > 0) and (h > 0)
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


def _is_xy_wh_s(bbox):
    """WARNING:
    Returns True if the bbox has the xy_XY format
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y), (w, h), s = bbox
        return (x >= 0) and (y >= 0) and (w > 0) and (h > 0) and 0 <= s <= 1
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


def _is_xy_XY(bbox):
    """WARNING:
    Returns True if the bbox has the xy_wh format and x < w and y < h
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y), (X, Y) = bbox
        return (0 <= x < X) and (0 <= y < Y)
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


def _is_xy_XY_s(bbox):
    """WARNING:
    Returns True if the bbox has the xywh format and x < w and y < h
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y), (X, Y), s = bbox
        return (0 <= x < X) and (0 <= y < Y) and 0 <= s <= 1
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


def _is_xywh(bbox):
    """WARNING:
    Returns True if the bbox has the xyXY format
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y, w, h) = bbox
        return (x >= 0) and (y >= 0) and (w > 0) and (h > 0)
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


def _is_xywh_s(bbox):
    """WARNING:
    Returns True if the bbox has the xyXY format
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y, w, h), s = bbox
        return (x >= 0) and (y >= 0) and (w > 0) and (h > 0) and 0 <= s <= 1
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


def _is_xywhs(bbox):
    """WARNING:
    Returns True if the bbox has the xyXY format
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y, w, h, s) = bbox
        return (x >= 0) and (y >= 0) and (w > 0) and (h > 0) and 0 <= s <= 1
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


def _is_xyXY(bbox):
    """WARNING:
    Returns True if the bbox has the xywh format and x < w and y < h
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y, X, Y) = bbox
        return (0 <= x < X) and (0 <= y < Y)
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


def _is_xyXY_s(bbox):
    """WARNING:
    Returns True if the bbox has the xywh format and x < w and y < h
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y, X, Y), s = bbox
        return (0 <= x < X) and (0 <= y < Y) and 0 <= s <= 1
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


def _is_xyXYs(bbox):
    """WARNING:
    Returns True if the bbox has the xywh format and x < w and y < h
    There's no way to check against that case

    Kept hidden to avoid cluttering the namespace
    """
    if not is_sequence(bbox):
        raise TypeError(f'Expected a sequence, got {type(bbox)}')

    try:
        (x, y, X, Y, s) = bbox
        return (0 <= x < X) and (0 <= y < Y) and 0 <= s <= 1
    except ValueError as e:
        if str(e).startswith('not enough values to unpack'):
            return False
        elif str(e).startswith('too many values to unpack'):
            return False
        else:
            raise
    except TypeError as e:
        if str(e).startswith('cannot unpack non-iterable'):
            return False
        elif "not supported between instances of" in str(e):
            return False
        elif 'object is not iterable' in str(e):
            return False
        else:
            raise


# Define at the end of the file because it requires
# the functions to be already defined
_FORMATS = {
    'xyXY': _is_xyXY,
    'xyXY_s': _is_xyXY_s,
    'xyXYs': _is_xyXYs,
    'xy_XY': _is_xy_XY,
    'xy_XY_s': _is_xy_XY_s,
    'xy_wh': _is_xy_wh,
    'xy_wh_s': _is_xy_wh_s,
    'xywh': _is_xywh,
    'xywh_s': _is_xywh_s,
    'xywhs': _is_xywhs,
}
